

import os
import time
from datetime import datetime
import threading
import sqlite3 as lite
from kervi.core.utility.thread import KerviThread
from kervi.plugin.storage.storage_plugin import StoragePlugin

_DB_CREATE_SQL = """
        CREATE TABLE `log` (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `sourceId`	TEXT,
            `sourceName`	TEXT,
            `area`	TEXT,
            `data`	TEXT,
            `level`	INTEGER,
            `topic`	TEXT,
            `body`	TEXT,
            'ts'    `timeStamp`	REAL,
            `logType`	TEXT
        );
        CREATE TABLE "dynamicData" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `dynamicValue`	TEXT,
            `value`	TEXT,
            `timeStamp`	TEXT
        );
        CREATE INDEX `sensorindex` ON `dynamicData` (`dynamicValue` ,`timeStamp` );
        CREATE TABLE "settings" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `setting_group`	TEXT,
            `name`	TEXT,
            `value`	TEXT
        );
        CREATE INDEX `settingsindex` ON `settings` (`setting_group` ,`name` );
        CREATE TABLE "cron" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `job_id` TEXT,
            `component_id` TEXT
            `name`	TEXT,
            `job_parameters` TEXT
        );
        CREATE TABLE "cron_meta" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `cron_id`	INTEGER,
            `repeat_start` REAL NOT NULL,
            `repeat_end` REAL NOT NULL,
            `repeat_interval` varchar(255) NOT NULL,
            `repeat_year` varchar(255) NOT NULL,
            `repeat_month` varchar(255) NOT NULL,
            `repeat_day` varchar(255) NOT NULL,
            `repeat_week` varchar(255) NOT NULL,
            `repeat_weekday` varchar(255) NOT NULL
        );
        """


class SQLiteStoragePlugin(StoragePlugin):
    def __init__(self, name, config, storage_type, manager):
        StoragePlugin.__init__(self, name, config, manager)
        self._storage_type = storage_type
        if self._storage_type == "memory":
            self._connection = lite.connect(self.manager.config.application.id + "_mem.db", check_same_thread=False)
        else:
            self._connection = lite.connect(self.manager.config.application.id + ".db", check_same_thread=False)
        self._init_db()
        self._db_lock = threading.Lock()
        self._ts_start = datetime.utcnow()
        
    def get_default_config(self):
        return {
            "name": None    
        }
    
    def _init_db(self):
        cursor = self._connection.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables_exists = False
        tables = cursor.fetchall()
        for table in tables:
            tables_exists = True
            break

        if not tables_exists:
            self._execute_sql(_DB_CREATE_SQL)        
            

    def _execute_sql(self, sql):
        sql_commands = sql.split(';')

        for command in sql_commands:
            try:
                self._connection.execute(command)
            except lite.Error as msg:
                self.log_error("create db, Command skipped: {0}, command{1}", msg, command)
    
    @property
    def storage_type(self):
        return self._storage_type
    
    def store_value(self, value_id, value, persist=False):
        self._db_lock.acquire()
        try:
            cursor = self._connection.cursor()
            cursor.execute(
                "INSERT INTO dynamicData ('dynamicValue','value','timeStamp')  VALUES (?, ?, ?)",
                (value["id"], self.to_json(value["value"]) , value["timestamp"])
            )
            self._connection.commit()
        except lite.Error as er:
            self.log_error('error store dynamic data, persist{0} error:{1}', persist, er)
        finally:
            self._db_lock.release()

    def get_value_data(self, value, date_from, date_to, limit):
        
        result = []
        if date_from is None:
            date_from = self._ts_start
        elif isinstance(date_from, str):
            date_from = datetime.strptime(date_from,'%Y-%m-%dT%H:%M:%S.%fZ')

        if date_to is None:
            date_to = datetime.utcnow()
        elif isinstance(date_to, str):
            date_to = datetime.strptime(date_to,'%Y-%m-%dT%H:%M:%S.%fZ')

        self._db_lock.acquire()
        try:
            cur = self._connection.cursor()
            cur.execute(
                "select * from dynamicData where dynamicValue=? and timestamp >= Datetime(?) and timestamp < Datetime(?)",
                (value, date_from, date_to)
            )

            all_rows = cur.fetchall()
            for row in all_rows:
                result += [
                    {
                        "value":self.to_json(row[2]),
                        "ts": row[3]
                    }
                ]
        
        except lite.Error as er:
            self.log_error('error get dynamic data:{0}', er)
        finally:
                self._db_lock.release()
                
        return result

    def store_setting(self, group, name, value):
        setting = self._retrieve_setting_db(group, name)
        self._db_lock.acquire()
        try:
            json_value = self.to_json(value)
            cur = self._connection.cursor()
        
            if setting:
                cur.execute(
                    "update settings set value=? where id=?",
                    (json_value, setting["id"])
                )
            else:
                cur.execute(
                    "INSERT INTO settings ('setting_group','name','value')  VALUES (?, ?, ?)",
                    (group, name, json_value)
                )
            self._connection.commit()
        except lite.Error as er:
            self.log_error('error store settings data:{0}', er)
            
        except Exception as er:
            self.log_error('error store settings data:{0}', er)
        finally:
            self._db_lock.release()

    def retrieve_setting(self, group, name):
        setting = self._retrieve_setting_db(group, name)
        
        if setting:
            return setting["value"]

    def _retrieve_setting_db(self, group, name):
        self._db_lock.acquire()
        try:
            cur = self._connection.cursor()
            cur.execute(
                "select * from settings where setting_group=? and name=?",
                (group, name)
            )

            all_rows = cur.fetchall()
            if len(all_rows) > 0:
                value = None
                try:
                    value = self.from_json(all_rows[0][3])
                except Exception:
                    pass

                return {
                    "id": all_rows[0][0],
                    "group": all_rows[0][1],
                    "name": all_rows[0][2],
                    "value": value
                }
        finally:
            self._db_lock.release()

        return None
    
    def store_message(self, source_id, message_item):
        raise NotImplementedError

    def get_messages(self):
        raise NotImplementedError    






import json
import inspect

class _ObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "to_json"):
            return self.default(obj.to_json())
        elif hasattr(obj, "__dict__"):
            data = dict(
                (key, value)
                for key, value in inspect.getmembers(obj)
                if not key.startswith("__")
                and not inspect.isabstract(value)
                and not inspect.isbuiltin(value)
                and not inspect.isfunction(value)
                and not inspect.isgenerator(value)
                and not inspect.isgeneratorfunction(value)
                and not inspect.ismethod(value)
                and not inspect.ismethoddescriptor(value)
                and not inspect.isroutine(value)
            )
            return self.default(data)
        return obj

from kervi.plugin.kervi_plugin import KerviPlugin
class StoragePlugin(KerviPlugin):
    def __init__(self, name, config, manager):
        KerviPlugin.__init__(self, name, config, manager)
    
    @property
    def storage_type(self):
        return "persisted"
    
    def store_value(self, value_id, value, persist=False):
        raise NotImplementedError

    def get_value_data(self, date_from, date_to, limit):
        raise NotImplementedError

    def store_setting(self, group, name, value):
        raise NotImplementedError

    def retrieve_setting(self, group, name):
        raise NotImplementedError

    def store_message(self, source_id, message_item):
        raise NotImplementedError

    def get_messages(self):
        raise NotImplementedError    

    def log_error(self, message, *args):
        self.spine.log.error(message, *args)

    def to_json(self, data):
        return json.dumps(data, cls=_ObjectEncoder, ensure_ascii=False).encode('utf8')

    def from_json(self, json_data):
        return json.loads(json_data.decode('utf8'))    




from kervi.config import Configuration
#from kervi.actions import Actions
from kervi.spine import Spine
class _Messaging(object):
    def __init__(self, **kwargs):
        self.spine = kwargs.get("spine", Spine())

    def send_message(self, subject, **kwargs):
        #kwargs = dict(kwargs, run_async=True)
        #Actions["message_manager.send_message"](subject, **kwargs)
        self.spine.send_command("messageManagerSend", subject, **kwargs)

Messaging = _Messaging()


#MIT License
#Copyright (c) 2017 Tim Wentzlau

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import kervi.core.utility.process as process


class _KerviModuleLoader(process._KerviProcess):
    """ Private class that starts a separate process that loads a module in the Kervi application """
    def init_process(self, **kwargs):
        self.spine.log.verbose("load: %s", self.name)

        try:
            import kervi.hal as hal
            hal._load()
            __import__(self.name, fromlist=[''])
       
        except ImportError:
            self.spine.log.exception("module not found:{0}", self.name)
        except:
            self.spine.log.exception("error load module:{0}", self.name)

        self.spine.send_command("startThreads", local_only=True)
        self.spine.trigger_event(
            "moduleLoaded",
            self.name,
        )

    def terminate_process(self):
        pass

    def load_spine(self, process_id, spine_port, root_address = None, ip="127.0.0.1"):
        from kervi.plugin.message_bus.bus_manager import BusManager
        self._bus_manager = BusManager()
        self._bus_manager.load(process_id, spine_port, root_address, ip)
        return self._bus_manager.bus




#MIT License
#Copyright (c) 2017 Tim Wentzlau

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import zmq
import random
import sys
import time
import inspect
import threading
import json
import uuid
import queue
import datetime
import logging
from kervi.plugin.message_bus.zmq.named_lists import NamedLists
import kervi.utility.nethelper as nethelper
from  kervi.core.utility.kervi_logging import KerviLog
from kervi.config.configuration import _KerviConfig

_KERVI_COMMAND_ADDRESS = "inproc://kervi_commands"
_KERVI_QUERY_ADDRESS = "inproc://kervi_query"
_KERVI_QUERY_RESPONSE_ADDRESS = "inproc://kervi_query_response"
_KERVI_EVENT_ADDRESS = "inproc://kervi_events"

class _ObjectEncoder(json.JSONEncoder):
    def default(self, o):
        
        if o and isinstance(o, datetime.datetime):
           return o.strftime("%Y-%m-%dT%H:%M:%SZ")
        elif isinstance(o, _KerviConfig):
            return o.as_dict()
        else:
            return json.JSONEncoder.default(self, o)

class ProcessConnection:
    def __init__(self, bus, is_root=False):
        self.address = None
        self.process_id = None
        self.is_connected = False
        self.is_root_connection = is_root
        self.include_ping = False
        self._bus = bus
        self._signal_socket = self._bus._context.socket(zmq.PUB)
        self._lock = threading.Lock()
        self.last_ping = None
        #self._signal_socket.setsockopt(zmq.SNDHWM, 25)

    def connect(self, address):
        self.address = address
        self._signal_socket.connect(address)

    # def connect_to(self, address, process_id):
    #     signal_message = {"address": address, "processId": process_id}
    #     p = json.dumps(signal_message, ensure_ascii=False).encode('utf8')
    #     signal_tag = "signal:connect"
    #     self.send_package([signal_tag.encode(), p])

    def register(self, address, process_id):
        self.address = address
        self.process_id = process_id
        self._signal_socket.connect(address)

    def disconnect(self):
        if self._signal_socket:
            self._lock.acquire()
            try:
                self._signal_socket.setsockopt(zmq.LINGER, 0)
                self._signal_socket.close()
                self._signal_socket = None
            finally:
                self._lock.release()

    def ping(self):
        self.last_ping = time.time()

    @property
    def is_alive(self):
        if self.is_connected:
            if time.time() - self.last_ping < 2:
                return True
        return False

    def send_package(self, package):
        self._lock.acquire()
        try:
            if self._signal_socket:
                self._signal_socket.send_multipart(package)
        except Exception as ex:
            self._bus.log.exception("send package exception")
        finally:
            self._lock.release()

class ZMQPingThread(threading.Thread):
    def __init__(self, bus):
        threading.Thread.__init__(self, None, None, "ZMQPing")
        self._bus = bus
        self.daemon = True
        self._terminate = False

    def stop(self):
        self._terminate = True

    def run(self):
        while not self._terminate:
            self._bus._ping_connections()
            time.sleep(.5)

class ZMQQueryThread(threading.Thread):
    def __init__(self, bus, tag, message):
        threading.Thread.__init__(self, None, None, "ZMQQyery")
        self._bus = bus
        self._message = message
        self._tag = tag
        self.daemon = True
        self._terminate = False

    def stop(self):
        self._terminate = True

    def run(self):
        self._bus._handle_message(self._tag, self._message)

class ZMQHandlerThread(threading.Thread):
    def __init__(self, bus):
        threading.Thread.__init__(self, None, None, "ZMQHandler")
        self._bus = bus
        self.daemon = True
        self._terminate = False
        self._messages = queue.Queue()

    def add_message(self, tag, message):
        self._messages.put((tag, message))

    def stop(self):
        self._terminate = True

    def run(self):
        while not self._terminate:
            try:
                tag, message = self._messages.get(True, 1)
                self._bus._handle_message(tag, message)
                self._messages.task_done()
            except queue.Empty:
                pass

class ZMQMessageThread(threading.Thread):
    def __init__(self, bus, address, bind=False):
        threading.Thread.__init__(self, None, None, "ZMQMessage")
        self._bus = bus
        self._address = address
        self._bind = bind
        self.daemon = True
        self._terminate = False
        self._socket = self._bus._context.socket(zmq.SUB)
        #self._socket.setsockopt( zmq.LINGER, 0 )

    def register(self, tag):
        self._socket.setsockopt_string(zmq.SUBSCRIBE, tag)

    def unregister(self, tag):
        self._socket.setsockopt_string(zmq.UNSUBSCRIBE, tag)

    def stop(self):
        self._terminate = True

    def connect(self):
        if self._bind:
            self._socket.bind(self._address)
        else:
            self._socket.connect(self._address)

    def run(self):
        while not self._terminate:
            connection_message = None
            try:
                connection_message = self._socket.recv_multipart()
                [tag, json_message] = connection_message
                message = json.loads(json_message.decode('utf8'))
                if tag == b"signal:exit":
                    if not self._bind:
                        break
                elif tag == b"queryResponse":
                    self._bus.resolve_response(message)
                else:
                    self._bus._add_message(tag.decode('utf-8'), message)
            except zmq.ZMQError as e:
                if e.errno == zmq.EAGAIN:
                    time.sleep(.001)
                    pass
                elif e.errno == zmq.ETERM:
                    self._terminate = True
                else:
                    self._bus.log.exception("message zmq exception: %s %s %s", self._address, e, e.errno)
            except Exception as e:
                self.bus.log.exception("message exception: %s %s %s", self._address, e, connection_message)
        self._socket.close()

class ZMQBus():
    _context = None
    _handlers = NamedLists()
    _connections = []
    _event_sock = None
    _command_sock = None
    _query_sock = None
    log = None

    def __init__(self):
        pass

    def set_log(self, logName):
        self.log = KerviLog(logName)

    @property
    def root_gone(self):
        return not self._is_root and time.time() - self._last_ping > 10

    def reset_bus(self, process_id, signal_port, ip="127.0.0.1", root_address=None, event_port=None):
        self._handlers = NamedLists()
        self._process_id = process_id
        self._query_id_count = 0
        self._uuid_handler = uuid.uuid4().hex
        self._is_root = (root_address is None)
        self._root_address = root_address
        self._signal_address = "tcp://"+ ip +":" + str(signal_port)
        self._context = zmq.Context()
        self._response_events = []
        self._last_ping = time.time()
        self._connections_lock = threading.Lock()

        self._linked_handlers = []
        self._linked_response_handlers = []

        self._message_threads = []
        self._message_thread = 0
        self._message_threads_lock = threading.Lock()
        for i in range(5):
            self._message_threads += [ZMQHandlerThread(self)]

        self._root_event = None
        self._event_socket = self._context.socket(zmq.PUB)
        self._event_socket.bind(_KERVI_EVENT_ADDRESS)

        self._command_socket = self._context.socket(zmq.PUB)
        self._command_socket.bind(_KERVI_COMMAND_ADDRESS)

        self._query_socket = self._context.socket(zmq.PUB)
        self._query_socket.bind(_KERVI_QUERY_ADDRESS)

        self._message_handler = ZMQMessageThread(self, self._signal_address, True)
        self._query_handler = ZMQMessageThread(self, _KERVI_QUERY_ADDRESS)
        self._event_handler = ZMQMessageThread(self, _KERVI_EVENT_ADDRESS)
        self._command_handler = ZMQMessageThread(self, _KERVI_COMMAND_ADDRESS)

        self._register_handler("signal:ping", self._on_ping)
        self._message_handler.register("signal:ping")
        self._message_handler.register("signal:exit")
        
        self._message_handler.register("queryResponse")
        self._message_handler.register("query:")

        self._query_handler.register("query:")

        self._message_handler.register("signal:exit")
        self._query_handler.register("signal:exit")
        self._event_handler.register("signal:exit")
        self._command_handler.register("signal:exit")

        self._ping_thread = ZMQPingThread(self)
        self._command_lock = threading.Lock()
        self._event_lock = threading.Lock()
        self._query_lock = threading.Lock()

        self.register_query_handler("GetRoutingInfo", self._get_routing_info)

    def _get_routing_info(self):
        result = []
        ignore_topics = [
            "GetRoutingInfo", 
            "ping", 
            "terminateProcess", 
            "getProcessInfo", 
            "processReady", 
            "appReady", 
            "authorizationAllowAnonymousUser",
            "authorizationValidSessionHeader",
            "authorizationRemoveSession",
            "startWebSocket",
            "stopThreads"
        ]
        for tag  in self._handlers.get_list_names():
            tag_info = tag.split(":")
            tag_id = None
            if len(tag_info) == 3:
                tag_id = tag_info[2]
            
            if not tag_info[1] in ignore_topics:
                result.append({
                    "id": tag_id,
                    "direction": "in",
                    "topic_type": tag_info[0],
                    "topic": tag_info[1]
                })
        return result
    
    def _add_linked_handler(self, func, **kwargs):
        groups = kwargs.get("groups", None)
        scopes = kwargs.get("scopes", [])
        argspec = inspect.getargspec(func)
        self._linked_handlers.append((func, groups, scopes, argspec.keywords != None))


    def _add_linked_response_handler(self, handler):
        self._linked_response_handlers.append(handler)

    def _register_handler(self, tag, func, **kwargs):
        groups = kwargs.get("groups", None)
        scopes = kwargs.get("scopes", [])

        argspec = inspect.getargspec(func)
        self._handlers.add(tag, (func, groups, scopes, argspec.keywords != None))

    
    def _unregister_handler(self, tag, func, **kwargs):
        groups = kwargs.get("groups", None)
        scopes = kwargs.get("scopes", [])

        argspec = inspect.getargspec(func)
        self._handlers.remove(tag, (func, groups, scopes, argspec.keywords != None))

    
    def get_handler_info(self):
        return self._handlers.get_list_names()

    def get_connection_info(self):
        result = []
        for connection in self._connections:
            result += [{"process": connection.process_id, "address": connection.address}]
        return result

    def _add_message(self, tag, message):
        with self._message_threads_lock:
            if tag.startswith("query:"):
                query = ZMQQueryThread(self, tag, message)
                query.start()
            else:
                self._message_threads[self._message_thread].add_message(tag, message)
                self._message_thread += 1
                if self._message_thread >= len(self._message_threads):
                    self._message_thread = 0

    def _handle_message(self, tag, message):
        func_list = [] 
        func_list += self._linked_handlers
        functions = self._handlers.get_list_data(tag)
        if functions:
            func_list += functions
        
        if tag.startswith("event:"):
            tag_parts = tag.split(":")
            event = "event:" + tag_parts[1] +":"
            if event!=tag:
                functions = self._handlers.get_list_data(event)
                if functions:
                    func_list += functions

        result = []
        session = None
        session_groups = None
        message_scopes = None
        process_id = None

        if "process_id" in message:
            process_id = message["process_id"]

        if "session" in message:
            session = message["session"]
            if session:
                session_groups = session['groups']
            else:
                session_groups = None

        if "scope" in message:
            message_scopes = message["scope"]

        if "injected" in message:
            injected = message["injected"]
        else:
            injected = None

        message_args = []
        message_kwargs = dict()
        if "kwargs" in message:
            message_kwargs = message["kwargs"]

        if "id" in message:
            if tag.startswith("query:"):
                message_kwargs["query_id"] = message["id"]
            else:
                message_args += [message["id"]]

        response_address = None
        if "responseAddress" in message:
            response_address = message["responseAddress"]

        if "address" in message:
            message_args += [message["address"]]

        if "processId" in message:
            message_args += [message["processId"]]

        if "processList" in message:
            message_args += [message["processList"]]

        if "args" in message:
            message_args += message["args"]

        

        message_kwargs = dict(message_kwargs, injected=injected, session=session, topic_tag=tag, response_address=response_address, process_id=process_id)
        send_response = True
        try:
            if func_list:
                for func, groups, handler_scopes, has_keywords in func_list:
                    authorized = True
                    if session_groups != None and groups and len(groups) > 0:
                        for group in groups:
                            if group in session_groups:
                                break
                        else:
                            authorized = False

                    if message_scopes:
                        for message_scope in message_scopes:
                            if message_scope in handler_scopes:
                                break
                        else:
                            authorized = False

                    if authorized:
                        if not has_keywords:
                            sub_result = func(*message_args)
                            if sub_result and sub_result == "****no_response****":
                                send_response = False
                            if sub_result:
                                result += [sub_result]
                        else:
                            sub_result = func(*message_args, **message_kwargs)
                            if sub_result and sub_result == "****no_response****":
                                send_response = False
                            elif sub_result:
                                result += [sub_result]
            if len(result) == 1:
                result = result[0]
        except Exception as e:
            raise e

        
        if response_address and send_response:
            self.send_query_response(response_address, message["id"], result)
        return result

    def send_query_response(self, response_address, query_id, result):
        message = {"messageType":"queryResponse", "address": self._signal_address, "id":query_id, "response":result}
        if response_address == "inproc_query":
            self.resolve_response(message)
        else:
            self.send_connection_message(response_address, "queryResponse", message)

    def run(self):

        for message_thread in self._message_threads:
            message_thread.start()

        self._message_handler.connect()
        self._message_handler.start()
        #time.sleep(1)
        self._event_handler.connect()
        self._command_handler.connect()
        self._query_handler.connect()

        self._query_handler.start()
        self._event_handler.start()
        self._command_handler.start()

        if self._root_address:
            self.connect_to_root()

        self._ping_thread.start()
        if self._is_root:
            logger = logging.getLogger()
            self.log.verbose("IPC address: %s", self._signal_address)
    def stop(self):
        exit_tag = "signal:exit"
        package = [exit_tag.encode(), json.dumps({}, ensure_ascii=False, cls=_ObjectEncoder).encode('utf8')]

        for connection in self._connections:
            connection.send_package(package)

        self._command_lock.acquire()
        try:
            self._command_socket.send_multipart(package)
        finally:
            self._command_lock.release()

        self._query_lock.acquire()
        try:
            self._query_socket.send_multipart(package)
        finally:
            self._query_lock.release()

        self._event_lock.acquire()
        try:
            self._event_socket.send_multipart(package)
        finally:
            self._event_lock.release()

        time.sleep(1)
        self._ping_thread.stop()
        self._ping_thread.join()
        
        self._message_handler.stop()
        self._event_handler.stop()

        self._command_handler.stop()
        self._query_handler.stop()
        
        for connection in self._connections:
            connection.disconnect()

        for message_thread in self._message_threads:
            message_thread.stop()
        
    def connect_to_root(self):
        self._root_event = threading.Event()
        connection = ProcessConnection(self, True)
        connection.process_id = "kervi-main"
        self._connections_lock.acquire()
        self._connections += [connection]
        connection.connect(self._root_address)
        self._connections_lock.release()

    def wait_for_root(self, timeout=20):
        if self._root_event:
            if self._root_event.wait(timeout):
                pass
            else:
                self.log.error("root not found! Process= %s address= %s", self._process_id, self._root_address)
            self._root_event = None

    @property
    def is_connected(self):
        result = True
        self._connections_lock.acquire()
        try:
            if not self._is_root and not self._connections:
                result = False

            if result:
                for connection in self._connections:
                    if not connection.is_connected:
                        result = False
        finally:
            self._connections_lock.release()
        return result

    def send_connection_message(self, address, tag, message):
        for connection in self._connections:
            if connection.address == address:
                p = json.dumps(message, ensure_ascii=False, cls=_ObjectEncoder).encode('utf8')
                connection.send_package([tag.encode(), p])
                return
        self.log.warn("connection not found %s %s %s", address, tag, message)

    def _on_ping(self, address, process_id, process_list):
        self._connections_lock.acquire()
        new_connection = True
        connection_list = []
        try:
            is_connected = False
            for process in process_list:
                if process["processId"] == self._process_id:
                    is_connected = True
                    break
            
            for connection in self._connections:
                if connection.process_id == process_id:
                    if not connection.is_connected and is_connected:
                        connection.is_connected = True
                        connection.include_ping = True
                        if connection.is_root_connection and self._root_event:
                            self._root_event.set()
                    if connection.is_connected:
                        connection.ping()

                    new_connection = False

            if new_connection:
                connection = ProcessConnection(self)
                connection.include_ping = True
                self._connections += [connection]
                connection.register(address, process_id)

            if not self._is_root and address == self._root_address:
                self._last_ping = time.time()
                for process in process_list:
                    if process["processId"] == self._process_id:
                        break
                    for connection in self._connections:
                        if connection.process_id == process["processId"]:
                            break
                    else:
                        connection = ProcessConnection(self)
                        self._connections += [connection]
                        connection.register(process["address"], process["processId"])

        finally:
            self._connections_lock.release()

    def _ping_connections(self):
        self._connections_lock.acquire()
        try:
            connection_list = []
            for connection in self._connections:
                if connection.include_ping:
                    connection_list += [{"address":connection.address, "processId":connection.process_id}]

            ping_message = {
                'address':self._signal_address,
                'processId':self._process_id,
                'processList': connection_list
            }
            p = json.dumps(ping_message, ensure_ascii=False, cls=_ObjectEncoder).encode('utf8')
            ping_tag = "signal:ping"
            package = [ping_tag.encode(), p]

            for connection in self._connections:
                connection.send_package(package)
        
        finally:
            self._connections_lock.release()

    def send_command(self, command, *args, **kwargs):
        injected = kwargs.pop("injected", "")
        scope = kwargs.pop("scope", None)
        groups = kwargs.pop("groups", None)
        session = kwargs.pop("session", None)
        local_only = kwargs.pop("local_only", False)
        command_message = {
            "command":command,
            "args":args,
            "injected":injected,
            "scope":scope,
            "session":session,
            "groups": groups,
            "kwargs": kwargs
        }
        p = json.dumps(command_message, ensure_ascii=False, cls=_ObjectEncoder).encode('utf8')
        command_tag = "command:" + command
        package = [command_tag.encode(), p]
        self._command_lock.acquire()
        try:
            self._command_socket.send_multipart(package)
        finally:
            self._command_lock.release()

        if not local_only:
            for connection in self._connections:
                connection.send_package(package)

    def register_command_handler(self, command, func, **kwargs):
        tag = "command:"+command
        self._register_handler(tag, func, **kwargs)
        self._command_handler.register(tag)
        self._message_handler.register(tag)

    def unregister_command_handler(self, command, func, **kwargs):
        tag = "command:"+command
        self._unregister_handler(tag, func, **kwargs)
        self._command_handler.unregister(tag)
        self._message_handler.unregister(tag)

    def trigger_event(self, event, id, *args, **kwargs):
        injected = kwargs.pop("injected", "")
        scope = kwargs.pop("scope", None)
        groups = kwargs.pop("groups", None)
        session = kwargs.pop("session", None)
        local_only = kwargs.pop("local_only", False)
        event_message = {
            'event':event,
            'id':id,
            'args':args,
            "injected":injected,
            "scope":scope,
            "groups":groups,
            "session":session,
            "kwargs": kwargs,
            "process_id": self._process_id
        }
        p = json.dumps(event_message, ensure_ascii=False, cls=_ObjectEncoder).encode('utf8')
        event_tag = "event:" + event + ":"
        if id:
            event_tag += id
        package = [event_tag.encode(), p]
        self._event_lock.acquire()
        try:
            self._event_socket.send_multipart(package)
        finally:
            self._event_lock.release()

        if not local_only:
            for connection in self._connections:
                connection.send_package(package)

    def register_event_handler(self, event, func, component_id=None, **kwargs):
        tag = "event:"+event +":"
        if component_id:
            tag +=  component_id
        if func:
            self._register_handler(tag, func, **kwargs)
        self._event_handler.register(tag)
        self._message_handler.register(tag)

    def unregister_event_handler(self, event, func, component_id=None, **kwargs):
        tag = "event:"+event +":"
        if component_id:
            tag +=  component_id
        
        if func:
            self._unregister_handler(tag, func, **kwargs)
        self._event_handler.unregister(tag)
        self._message_handler.unregister(tag)

    def resolve_response(self, message):
        for event in self._response_events:
            if event["id"] == message["id"] and not event["processed"]:
                if message["response"]:
                    event["response"] += [message["response"]]
                event["process_count"] = event["process_count"] - 1
                event["handled_by"] += [message["address"]]
                if  event["process_count"] <= 0:
                    event["processed"] = True
                    event["eventSignal"].set()
                    for handler in self._linked_response_handlers:
                        handler(event)

    def send_query(self, query, *args, **kwargs):
        self._connections_lock.acquire()
        try:
            self._query_id_count += 1
            result = []
            injected = kwargs.pop("injected", "")
            scope = kwargs.pop("scope", None)
            groups = kwargs.pop("groups", None)
            session = kwargs.pop("session", None)
            processes = kwargs.pop("processes", None)
            timeout = kwargs.pop("timeout", 10)
            wait = kwargs.pop("wait", True)
            headers = kwargs.pop("headers", None)
            local_only = kwargs.pop("local_only", False)
            query_id = self._uuid_handler + "-" + str(self._query_id_count)

            process_count = 1
            if not local_only:
                for connection in self._connections:
                    if connection.is_alive and (not processes or (connection.process_id in processes)):
                        process_count += 1

            event = threading.Event()
            event_data = {
                "id":query_id,
                "eventSignal":event,
                "response":[],
                "processed":False,
                "process_count": process_count,
                "process_id": self._process_id,
                "query": query,
                "handled_by": [],
                "headers": headers
            }
            self._response_events += [event_data]
            query_message = {
                'query':query,
                "id":query_id,
                "responseAddress":"inproc_query",
                'args':args,
                "injected":injected,
                "scope":scope,
                "groups":groups,
                "session":session,
                "kwargs": kwargs
            }
            p = json.dumps(query_message, ensure_ascii=False, cls=_ObjectEncoder).encode('utf8')
            query_tag = "query:" + query
            package = [query_tag.encode(), p]
            self._query_lock.acquire()
            try:
                self._query_socket.send_multipart(package)
            finally:
                self._query_lock.release()

            if not local_only and len(self._connections) > 0:
                query_message["responseAddress"] = self._signal_address
                p = json.dumps(query_message, ensure_ascii=False, cls=_ObjectEncoder).encode('utf8')
                package = [query_tag.encode(), p]

                for connection in self._connections:
                    if connection.is_alive and (not processes or (connection.process_id in processes)):
                        connection.send_package(package)
        except Exception as ex:
            self.log.exception("error send query %s", query)
        finally:
            self._connections_lock.release()

        if wait:
            event.wait(timeout)
            if not event_data["processed"]:
                self.log.warn("send query timeout %s %s %s %s", self._signal_address, query, event_data["handled_by"], event_data["id"])
            result = event_data["response"]
            self._response_events.remove(event_data)
            if isinstance(result, list) and not isinstance(result, dict) and len(result) == 1:
                return result[0]
            else:
                return result
        return None

    def register_query_handler(self, query, func, **kwargs):
        self._register_handler("query:"+query, func, **kwargs)

    def unregister_query_handler(self, query, func, **kwargs):
        self._unregister_handler("query:"+query, func, **kwargs)



#MIT License
#Copyright (c) 2017 Tim Wentzlau

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module that holds classes for creating / bootstraping a Kervi application or a Kervi module.
"""
import os
import platform
if platform.system() == "Windows":
    os.system('color')
import time
import threading
import logging
from kervi.core.utility.kervi_logging import KerviLog

try:
    import thread
except ImportError:
    import _thread as thread

import sys
if sys.version_info[0] < 3:
    print("Kervi requires Python 3")
    exit()

from kervi.application.kervi_module import KerviModule
import kervi.utility.nethelper as nethelper
from kervi.controllers import Controller
from kervi.actions import action
from kervi.application.default_config import get_default_config
from kervi.plugin.plugin_manager import PluginManager
from kervi.utility.discovery import KerviAppDiscovery
from kervi.utility.logging_handler import KerviLogHandler
import logging
import signal

_app_running = True
_console_log = 15

def handler_stop_signals(signum, frame):
    global _app_running
    _app_running = False
    logging.getLogger().fatal("signal: %s", signum)

def _pretty_print(d, indent=0):
    if isinstance(d, dict):
        for key in d.keys():
            value = d[key]
            
            if isinstance(value, dict) or isinstance(value, list):
                print( '  ' * indent + str(key))
                _pretty_print(value, indent+1)
            else:
                print('  ' * (indent+1) + key + ":" +  str(value))
    elif isinstance(d, list):
        for item in d:
            if isinstance(item, dict) or isinstance(item, list):
                _pretty_print(item, indent+1)
            else:
                print('  ' * (indent+1) + str(item))
    else:
        pass    




class _AppActions(Controller):
    def __init__(self, app):
        super().__init__("app", "App controller")

        self._app = app

    @action(confirm=True, confirm_message="Shutdown device")
    def shutdown(self):
        self._app._shutdown_action()

    @action(confirm=True, confirm_message="Reboot device")
    def reboot(self):
        self._app._reboot_action()

    @action(confirm=True, confirm_message="Restart application")
    def restart(self):
        self._app._restart_action()

    @action(confirm=True, confirm_message="Stop application")
    def stop(self):
        self._app._stop_action()

class Application(object):
    """ Kervi application class that starts a kervi application and loads all needed modules.
        This class should be used by it self like:

    """
    def __init__(self, user_config = None):
        """ Settings is a dictionary with the following content
        """
        global _console_log
        
        print("\033[92mStarting kervi application \033[0m")
        import inspect
        import getopt
        self._websocket_event = threading.Event()
        self._in_stop = False
        self._discovery_thread = None
        config_files = []
        self._webserver_info = None
        opts, args = getopt.getopt(sys.argv[1:], "c", ["config_file=", "as-service", "install-service", "uninstall-service", "start-service", "stop-service", "restart-service", "status-service", "detect-devices"])
        for opt, arg in opts:
            if opt in ("-c", "--config_file"):
                if os.path.isfile(arg):
                    config_files += [arg]
                else:
                    print("Specified config file not found:", arg)
        script_path = os.path.abspath(inspect.stack()[1][1])
        script_name = os.path.basename(script_path)
        script_name, script_ext = os.path.splitext(script_name)
        
        config_files += [script_name +".config.json"]
        config_files += ["kervi.config.json"]

        selected_config_file = None
        for config_file in config_files:
            if os.path.isfile(config_file):
                selected_config_file = config_file
                break

        from kervi.config import load
        self.config = load(
            config_file=selected_config_file,
            config_user=user_config,
            config_base=get_default_config()
        )

        service_commands = []
        detect_devices = None
        self._as_service = False
        for opt, arg in opts:
            if opt in ("--as-service"):
                self._as_service = True

            if opt in ("--detect-devices"):
                detect_devices = True

            if opt in ("--install-service"):
                service_commands += ["install"]

            if opt in ("--uninstall-service"):
                service_commands += ["uninstall"]

            if opt in ("--start-service"):
                service_commands += ["start"]

            if opt in ("--stop-service"):
                service_commands = ["stop"]
            
            if opt in ("--restart-service"):
                service_commands = ["restart"]
            if opt in ("--service-status"):
                service_commands = ["status"]

        if service_commands:
            import kervi.hal as hal
            hal_driver = hal._load(self.config.platform.driver)
            if hal_driver:
                hal.service_commands(
                    service_commands,
                    self.config.application.name,
                    self.config.application.id,
                    script_path
                )
            exit()

        if detect_devices:
            import kervi.hal as hal
            hal_driver = hal._load(self.config.platform.driver)
            if hal_driver:
                devices = hal.detect_devices()
                print("devices:")
                _pretty_print(devices)
            exit()
        self._log_handler = KerviLogHandler(self.config)
        self._log_queue = self._log_handler._log_queue 
        self._logger = KerviLog("application")
        #self._validateSettings()
        self.started = False
        self._webserver_port = None

        try:
            from kervi.version import VERSION
        except:
            VERSION="0.0.0"

        self._logger.verbose("kervi version: %s", VERSION)

        import kervi.hal as hal
        hal_driver = hal._load(self.config.platform.driver)
        if hal_driver:
            self._logger.verbose("platform driver: %s", hal_driver)
            self._logger.verbose("board: %s", hal.get_board_name())
            
        from kervi.plugin.message_bus.bus_manager import BusManager
        self.bus_manager = BusManager(self._log_queue)
        self.config.network.ipc_root_port = nethelper.get_free_port([self.config.network.ipc_root_port])
        self.bus_manager.load("kervi-main", self.config.network.ipc_root_port, None, self.config.network.ipc_root_address)
        
        from kervi import spine
        self.spine = spine.Spine()
        self.spine.register_query_handler("GetApplicationInfo", self._get_application_info)
        self.spine.register_query_handler("getProcessInfo", self.get_process_info)
        self.spine.register_event_handler("modulePing", self._module_ping)
        self.spine.register_event_handler("processReady", self._process_ready, scope="app-" + self.config.application.id)
        self.spine.register_event_handler("WebAppReady", self._webapp_ready, scope="app-" + self.config.application.id)
        self.spine.register_event_handler("websocketReady", self._websocket_ready, scope="app-" + self.config.application.id)
        self._module_processes = []
        self._process_info = []
        self._process_info_lock = threading.Lock()

        self._kervi_modules = []
        self._kervi_modules_lock = threading.Lock()

        from kervi.storage.storage_manager import StorageManager
        self._storage_manager = StorageManager(self._log_queue)
    
        from kervi.utility.authorization_manager import AuthorizationManager
        self._authorization_manager = AuthorizationManager(self._log_queue)
        
        from kervi.messaging.message_manager import MessageManager
        self._message_manager = MessageManager(self._log_queue)

        self._app_actions = _AppActions(self)

    @property
    def actions(self):
        return self._app_actions

    def _validateSettings(self):
        pass

    def _get_application_info(self):
        unit_system = self.config.unit_system
        display = {}
        display["unit_system"] = self.config.display.unit_systems.systems[unit_system]
        display["texts"] = self.config.texts.ui

        return {
            "name": self.config.application.name,
            "id": self.config.application.id,
            "display": display,
            "unit_system": unit_system
        }

    def get_process_info(self):
        return {"id": "application"}

    def _module_ping(self, module_id):
        with self._kervi_modules_lock:
            for module in self._kervi_modules:
                if module.module_id == module_id:
                    
                    if not module.is_alive:
                        self._logger.verbose("module started: %s", module_id)
                        self.spine.trigger_event(
                            "moduleStarted",
                            module_id
                        )
                    module.ping()
                    break
            else:
                self._logger.verbose("module started: %s", module_id)
                self._kervi_modules += [KerviModule(module_id)]
                self.spine.trigger_event(
                    "moduleStarted",
                    module_id
                )

    def _check_kervi_modules(self):
        with self._kervi_modules_lock:
            for module in self._kervi_modules:
                if not module.is_alive and module.connected:
                    self._logger.verbose("module stopped: %s", module.module_id)
                    module.connected = False
                    self.spine.trigger_event(
                        "moduleStopped",
                        module.module_id
                    )

    def _process_ready(self, scope, process_id):
        self._process_info_lock.acquire()
        try:
            for process in self._process_info:
                if process["id"] == process_id:
                    process["ready"] = True
        finally:
            self._process_info_lock.release()

    def _webapp_ready(self, scope, webserver_info):
        self._webserver_info = webserver_info
        ready_message = "Reach your application at http://" + self._webserver_info["ip"] + ":" + str(self._webserver_info["port"])
        self.spine.log.info(ready_message)
        self._websocket_event.wait()
        self.spine.send_command("startWebSocket")

    def _websocket_ready(self, scope, info):
        self._websocket_event.set()
    
    def _is_ready(self):
        result = True
        self._process_info_lock.acquire()
        try:
            for process in self._process_info:
                if not process["ready"]:
                    result = False
        finally:
            self._process_info_lock.release()

        return result

    def _start(self):
        self.started = True

        try:
            import dashboards
        except ImportError:
            pass
        
        import kervi.core.utility.process as process
        import kervi.utility.application_helpers as app_helpers


        #if self._as_service:
        signal.signal(signal.SIGINT, handler_stop_signals)
        signal.signal(signal.SIGTERM, handler_stop_signals)

        self.spine.send_command("startThreads", local_only=True)
        time.sleep(.5)
        module_port = self.config.network.ipc_root_port
        pluginManager = PluginManager(self.config, log_queue=self._log_queue)
        self._process_info_lock.acquire()
        plugin_modules = pluginManager.prepare_load()
        for plugin_module in plugin_modules:
            self._process_info.append(
                {"id":plugin_module, "ready": False}
            )
        self._process_info_lock.release()

        module_port = pluginManager.load_plugins(module_port+1)

        for module in self.config.modules:
            self._process_info_lock.acquire()
            self._process_info += [{"id":module, "ready":False}]
            self._process_info_lock.release()

            module_port += 1
            self._module_processes += [
                process._start_process(
                    "app-" + self.config.application.id,
                    module,
                    self.config,
                    nethelper.get_free_port([module_port]),
                    app_helpers._KerviModuleLoader,
                    log_queue = self._log_queue
                )
            ]

        #print("wait for ready")
        while not self._is_ready():
            time.sleep(1)

        #print("pi", self._process_info)

        if not self._in_stop:
            from kervi.dashboards import Dashboard
            Dashboard._add_default()

            self.spine.send_command("kervi_action_app_main")
            ready_message = "Your Kervi application is running"
            self._logger.info(ready_message)

            self._logger.info("Press ctrl + c to stop your application")
            self.spine.trigger_event(
                "appReady",
                self.config.application.id
            )

            if self.config.discovery.enabled:
                self._discovery_thread = KerviAppDiscovery(
                    self.config.network.ip, 
                    self.config.network.ipc_root_port,
                    self.config.discovery.port, 
                    self.config.application.id, 
                    self.config.discovery.challenge,
                    self.config.application.name,
                    "http://" + self.config.network.ip# + ":" + str(self.config.network.http_port)
                )
                self._discovery_thread.start()
            else:
                self._discovery_thread = None
            
    def _input_thread(self, list):
        try:
            raw_input()
            list.append(None)
        except KeyboardInterrupt:
            pass
        except:
            pass

    def run(self):
        """Run the application loop. The application will continue until termination with ctrl-c"""

        if not self.started:
            self._start()
        try:
            char_list = []
            thread.start_new_thread(self._input_thread, (char_list,))
            while not char_list and _app_running:
                self._check_kervi_modules()
                time.sleep(1)

        except KeyboardInterrupt:
            pass

        self.stop()

    def _xrun(self):
        """Used in tests"""

        if not self.started:
            self._start()

    def stop(self, force_exit=True, restart=False):
        self._logger.warn("stopping processes")
        self._in_stop = True
        self.spine.send_command("kervi_action_app_exit")
        
        if self._discovery_thread:
            self._discovery_thread.terminate()

        import kervi.core.utility.process as process
        process._stop_processes("app-" + self.config.application.id)
        self.spine.trigger_event("processTerminating", None, local_only=True)
        time.sleep(1)
        self.bus_manager.stop()
        self._log_handler.stop()
        #time.sleep(1)
        #for thread in threading.enumerate():
        #    print("running thread",thread.name)
        self._logger.info("application stopped")
        #import psutil
        #current_process = psutil.Process()
        #children = current_process.children(recursive=True)
        #for child in children:
        #    print('Child pid is {}'.format(child.pid))

        if force_exit:
            import os
            os._exit(0)

    def _stop_action(self):
        self._logger.verbose("stop action","")
        self.stop()

    def _restart_action(self):
        self._logger.verbose("restart action","")
        self.stop(restart=True)

    def _reboot_action(self):
        self._logger.verbose("reboot action","")
        import kervi.hal as hal
        hal.device_reboot()

    def _shutdown_action(self):
        self._logger.verbose("shutdown action","")
        import kervi.hal as hal
        hal.device_shutdown()
